
#include "ThreadGlobals.hpp"
#include "Satistics.hpp"
#include <signal.h>

#include "Types.hpp"
#include "IFStream.hpp"
#include "Logging.hpp"
#include "AssignInfos.hpp"
#include "BasicSolver.hpp"
#include <string>
#include <algorithm>
#include <cctype>
#include "Assertion.hpp"
//#include "DMRPSolver.hpp"

using std::string;

const char*  main_cpp_version(){return "$Revision: 1.16 $";}


/**
 * Some checks
 */

bool check_defines(){

	std::cout << "c asrts will be evaluated and wordsize is " << sizeof(void*) << std::endl;
    return
    //sizeof(void*)           ==      32/8                &&
    sizeof(1)				==		32/8				&&
    sizeof(int32_t)         ==      32/8                &&
    sizeof(int64_t)         ==      64/8                &&
    sizeof(uint32_t)        ==      32/8                &&
    sizeof(uint64_t)        ==      64/8                &&
    sizeof(double)          ==      64/8                &&
    sizeof(float)           ==      32/8                &&
    sizeof(Literal)         ==      sizeof(uint32_t)    &&
    sizeof(Variable)        ==      sizeof(uint32_t)    &&
    sizeof(float)           ==      sizeof(uint32_t)    &&
    ((int)false)            ==      0                   &&
    ((int)true)             ==      1;
}


LOGGING_INIT("./log_solver.csv")
//ASRT_INIT


TGlobals					glob;

static void SIGINT_handler(int signum) {
	string what="x";
	do{
		std::transform(what.begin(), what.end(), what.begin(),tolower);
		if(what=="k")	{ exit(0);}
		if(what=="q")	{ glob.sets.timeout = 0; break;}
		if(what=="p")	{ glob.stats.printStatsConsole(UNKNOWN_TrBool);}
		if(what=="c")	{ break;}
		if(what=="o")	{ glob.sets.printUsage();}
		if(what=="r")	{ std::cin >> what; glob.sets.read(what.c_str());}


		cout << endl << "What to do?" << endl;
		cout << "k:       kill immediately" << endl;
		cout << "q:       quit after next restart" << endl;
		cout << "p:       print statistics" << endl;
		cout << "o:       show options and current option setting" << endl;
		cout << "r <opt>: read (non initial) option with new value (e.g. r -pol-mode=1)" << endl;
		cout << "c:       continue" << endl;
	}while(std::cin >> what);
	cout << "go ahead ... " << endl;
}

void copyCNFClses(string from, string to, uint32_t vars, vector<uint32_t>& ids){

	std::sort(ids.begin(),ids.end());

	std::ofstream out(to.c_str());
	if(out.bad()){
		cout << "c Could not open Out file " <<  to << endl; return;}

	IFStream in(from.c_str());
	if(in.isBad()){
	    cout << "c Could not open CNF file " <<  from << endl; return;}

	out << "p cnf " << vars << " " << ids.size() << endl;

    uint32_t curr_id = 0; uint32_t p_idx = 0; int lit;
  	in.skipWhitespace();
   	while(in.hasNext()){
   		if(in.currEquals('c') || in.currEquals('p')) {in.skipLine();continue;}

   		if(p_idx>=ids.size()) break;

   		++curr_id; 		asrt(ids[p_idx] >= curr_id);

   		const bool take = ids[p_idx] == curr_id;

   		while((lit = in.getNextInt()) != 0){
   			if(take){out << lit << " ";}
   		}
   		if(take){out << "0" << endl; p_idx++;}
   	}
}



void copyGCNFClses(string from, string to,vector<uint32_t>& ids){

	const bool print_as_gcnf = false; //if true a "p gcnf" file is printed otherwise a normal cnf file

	vector<bool> print;
	print.push_back(true);
	print.resize(ids.size(),false);
	for(uint32_t i=0;i!= ids.size();++i){
		uint32_t grp = ids[i];
		if(grp >= print.size()){print.resize(grp + 1,false);}
		print[grp] = true;
	}


	std::ofstream out(to.c_str());
	if(out.bad()){
		cout << "c Could not open Out file " <<  to << endl; return;}

	IFStream in(from.c_str());
    if(in.isBad()){
	    cout << "c Could not open CNF file " <<  from << endl; return;}

    //header
	if(print_as_gcnf){	out << "p gcnf ";}
	else{				out << "p cnf ";}
	std::streampos pline = out.tellp();
	out << "                                                       " << endl;

	uint32_t cls 		= 0; int32_t grp, lit;
	uint32_t max_var 	= 0;
	uint32_t max_grp 	= 0;
   	while(in.hasNext()){
   		if(in.currEquals('c') || in.currEquals('p')) {in.skipLine();continue;}

   		if(in.currEquals('{')){
   			++in;
   			grp = in.getNextInt();
   			in.skipWhitespace();
   		}else{grp = -1;}

   		if(grp < 0 || !in.currEquals('}')){
   			cout << "c PARSE ERROR in file " << from << endl,
   			exit(EXIT_FAILURE);
   		}
   		++in;
   		const bool take = grp < (int32_t)print.size() && print[grp];

   		if(take && print_as_gcnf){
   			out << "{" << grp << "} ";
   			if((uint32_t)grp > max_grp){max_grp = grp;}
   		}

   		while((lit = in.getNextInt()) != 0){
   			if(take){
   				out << lit << " ";
   				uint32_t abs = lit > 0 ? lit : -lit;
   				if(abs > max_var){max_var = abs;}
   			}
   		}
   		if(take){out << "0" << endl; ++cls;}
   	}

   	out.seekp(pline);
   	out << max_var << " " << cls;
   	if(print_as_gcnf){out << " " << max_grp;}
}



int main(int argc, char** argv){

	BasicSolver			solver(glob);
	Settings& sets		= glob.sets;
	Satistics& stats	= glob.stats;

	sanity((printf("sanity checks will be evaluated\n"),true));
    asrt(check_defines()); asrt(solver.vals.litTrue(TRUE_LIT));

    //LOGGING_OFF ASRT_OFF

    if(!sets.read(argc,argv)){
        sets.printUsage();
        return EXIT_FAILURE;
    }
    glob.randy.setSeed(sets.rand_seed);

    //try to open file
    IFStream in(sets.cnf_file_name.c_str()); asrt(solver.vals.litTrue(TRUE_LIT));
    //FILE * in = fopen (sets.cnf_file_name, "r");
    if(in.isBad()){
        cout << "c Could not open CNF file " <<  sets.cnf_file_name << endl;
        return EXIT_FAILURE;
    }

    signal(SIGINT,SIGINT_handler);
    //signal(SIGQUIT,SIGQUIT_handler);

    if(sets.print_console ){stats.printProgressHead();}

    T_ON(parse); asrt(solver.vals.litTrue(TRUE_LIT));
    bool ok=solver.parseDIMACS(in);
    in.close();
    T_OFF(parse);

    glob.stats.keepOrig(solver.vals.count()); asrt(solver.vals.litTrue(TRUE_LIT));



    asrt(solver.vals.litTrue(TRUE_LIT));

    //TrBool result = ok ? UNKNOWN_TrBool : FALSE_TrBool;
    TrBool result = ok ? solver.isSAT() : FALSE_TrBool;


    if(result.isKnown()){
		if(result.isTrue()){
			cout << endl << "s SATISFIABLE" << endl;

			if(sets.print_values) {solver.printModel(cout);}

			if(!sets.out_file_name.empty()){
				std::ofstream out(sets.out_file_name.c_str());
				solver.printModel(out);}

			/*if(!solver.asrtVerifyModel(sets.cnf_file_name.c_str())){
				LOG0("Model could not be validated");}*/
		}else{
			cout << endl << "s UNSATISFIABLE" << endl;

			if(sets.minimise){
				solver.minimiseCore();}
			else{
				char* why =  glob.whyInvalid();
				if(why != NULL){cout << "c Reason for Unsat: " << why << endl;}
			}

			if(sets.print_values){solver.printCore(NULL);}

			if(!sets.out_file_name.empty()){
				vector<uint32_t> cls_ids;
				solver.printCore(&cls_ids);
				if(solver.cls_mode){
					copyCNFClses(sets.cnf_file_name,sets.out_file_name,solver.sel_strt.idx() - 1, cls_ids);
				}else{
					copyGCNFClses(sets.cnf_file_name,sets.out_file_name,cls_ids);
				}
			}
		}
   }else{
	   printf("s UNKNOWN\n");
   }

    if(!sets.csv_file_name.empty()){
        OFStream out(sets.csv_file_name.c_str(),false);
        stats.printCSV(sets.csv_file_name.c_str(),result);}


    if(sets.print_console){
    	stats.printStatsConsole(result);
    }


    uint32_t ret_val = result.isKnown() ? (result.isTrue()?10:20) : 0;
#ifdef NDEBUG
	exit(ret_val);
#endif
    return ret_val;
}







